class Ball
{
  float xpos;
  float ypos;
  float speed;        // in pixels per frame
  float direction;    // in radians
  float diameter;
  color colour;

  int invincCounter;                                      // time a ball will turn invincible when hit; in ms

  int[] soundCounters = new int[bpmeasure * tpb];         // how many times sound has sounded while on
  boolean[] soundSounded = new boolean[bpmeasure * tpb];  // if sound already played for current tick
  boolean[] on = new boolean[bpmeasure * tpb];            // if activated for current tick
  AudioSnippet audio;                                     //  to play the ball's sound

  Ball(String af, color c)
  {
    speed = random(4.0) + 2.0;
    direction = random(TWO_PI);
    diameter = random(12.0) + 12.0;
    reposition();
    colour = c;
    audio = Minim.loadSnippet(af);
  }

  void reposition()
  {
    xpos = random(width - 2 * diameter) + diameter;
    ypos = random(height - 2 * diameter - infoBarHeight) + diameter;
  }
  
  void move()
  {
    xpos = xpos + cos(direction) * speed;
    ypos = ypos + sin(direction) * speed;
    float r = 0;//diameter/2;
    if(xpos < r)
    {
      xpos = width - r;
    }
    else if(xpos > width - r)
    {
      xpos = r;
    }

    if(ypos < r)
    {
      ypos = height - infoBarHeight - r;
    }
    else if(ypos > height - infoBarHeight - r)
    {
      ypos = r;
    }
    render();
  }

  void render()
  {
    fill(colour);
    stroke(colour);
    ellipse(xpos, ypos, diameter, diameter);
  }

  // the ball checks if it is hit by the player, and if true,
  // changes direction slightly, becomes invincible temporarily, and changes its sound pattern
  // currentRadius is the radius within which a hit will occur  
  void checkForPlayerHit(float currentRadius)
  {
    // see if the ball is currently invincible to the player
    boolean invinc;
    if(invincCounter > 0)
    {
      invincCounter -= millis() - previousmillis;
      invinc = true;
    }
    else
    {
      invinc = false;
    }
    // determine the direct distance to player
    float distance = dist(mouseX, mouseY, xpos, ypos) - diameter/2.0;

    if(!invinc && distance < currentRadius) 
    {
      // hit for the first time by player
      // switch sound on (or 'reset')
      on[currentTick] = true;
      soundCounters[currentTick] = 0;
      soundSounded[currentTick] = false;

      // turn invincible for one OR two OR three ticks (randomly)
      float r = random(1.0);
      if(r < 0.333)
      {
        invincCounter = mspt;
      }
      else if(r > 0.333 && r < 0.667)
      {
        invincCounter = 2 * mspt;
      }
      else
      {
        invincCounter = 3 * mspt;
      }      
    }
    else if(invinc)
    {
      // ball is invincible so was already hit for the first time;
      // change direction randomly every frame
      float directionChange = random(0.6) - 0.3;
      direction += directionChange;
    }  
  }

  // the ball checks if it is hit by another ball, and if true,
  // exchanges direction and speed with the other one
  boolean checkForBallCollision(Ball other)
  {
    float distance = dist(this.xpos, this.ypos, other.xpos, other.ypos);
    distance -= diameter / 2.0;
    distance -= other.diameter / 2.0;
    if(distance <= 0.0) 
    {
      //exchange speed values
      float tempspeed = other.speed;
      other.speed = this.speed;
      this.speed = tempspeed;

      // exchange direction values          
      float tempdirection = other.direction;
      other.direction = this.direction;
      this.direction = tempdirection;
      
      return true;
    }
    return false;
  }

  // the ball checks if it is time to play its sound
  // also, resets values for soundSounded
  void checkForSound()
  {
    if(on[currentTick])
    {
      // blink, also when sound sounded already in previous frame 
      // within this tick (to make it more visible)
      blink();
      // play its sound only when sound hasn't sounded already for this tick
      if(!soundSounded[currentTick])
      {
        audio.rewind();
        audio.play();
        soundSounded[currentTick] = true;
        soundCounters[currentTick]++;
        // stop making the sound if the limit is reached
        if(soundCounters[currentTick] > soundLimit)
        {
          on[currentTick] = false;
          soundCounters[currentTick] = 0;
        }
      }
    }

    // always set the soundSounded of former tick back to false
    if(currentTick == 0)
    {
      soundSounded[soundSounded.length - 1] = false;
    }
    else
    {
      soundSounded[currentTick - 1] = false;
    }
  }

  void blink()
  {
    // show up white for one frame
    stroke(100);
    fill(0);
    ellipse(xpos, ypos, 1.6*diameter, 1.6*diameter);
    stroke(255);
    ellipse(xpos, ypos, 1.3*diameter, 1.3*diameter);
    fill(255);
    ellipse(xpos, ypos, diameter, diameter);
    redraw();
  }    
}
