import processing.core.*; import processing.opengl.*; import ddf.minim.*; import java.applet.*; import java.awt.*; import java.awt.image.*; import java.awt.event.*; import java.io.*; import java.net.*; import java.text.*; import java.util.*; import java.util.zip.*; public class soundBalls3 extends PApplet {


// fps = frames per second
// bpm = beats per minute
// bps = beats per second
// bpmeasure = beats per measure
// tps = ticks per second
// fpt = frames per tick
// tpb = ticks per beat
int tpb = 3;
int bpmeasure = 4;
int mspt;                  // how many ms every tick takes
int bpminute;              // current beats per minute (tempo)
int currentTick = 0;       // which tick in the measure we are now
int millicounter = 0;      // how many ms passed since previous tick            
int previousmillis = 0;    // how many ms passed since previous frame            
int soundLimit = 11;       // number of measures balls play their sound once triggered

String fpstext = "INIT";
String bpmtext = "INIT";
int infoBarHeight = 70;    // height of bottom info bar
int fpsWidth = 190;        // weight of part of info bar showing FPS
int bpmWidth = 223;        // weight of part of info bar showing BPM

// the radiation radiated by the player
Radiation radiation = new Radiation();
// array of balls in the game (maximum of 10)
Ball[] balls = new Ball[10];
// array of pickups in the game
Pickup[] pickups = new Pickup[5];

AudioSnippet beatAudio;

public void setup()
{
  size(900, 650, OPENGL); // 900, 650
  noFill();
  smooth();
  noCursor();  

  // try to achieve 60 frames per second and start with 100 bpm
  frameRate(60);
  setBPM(100);

  textFont(loadFont("Pinball-Data-48.vlw"));
  textAlign(LEFT);
  rectMode(CORNER);
  strokeWeight(1);

  // initialize sound
  Minim.start(this);
  beatAudio = Minim.loadSnippet("button2.wav");
  // sound indicating beat set off by setting it to -80 dB
  beatAudio.setGain(-80);    //-6

  // spawn all Balls
  String audioFile;
  for(int c = 0; c < balls.length; c++)
  {
    audioFile = "c" + (c + 1) + ".wav";
    balls[c] = new Ball(audioFile, getColour(c + 1));
    boolean collision = true;
    while(collision)
    {
      for(int i = 0; i < c; i++)
      {
        if(i != c)
        {
          if(balls[c].checkForBallCollision(balls[i]))
          {
            collision = true;
            balls[c].reposition();
            break;
          }
        }
      }
      collision = false;
    }
  }

  // prepare all Pickups
  pickups[0] = new Pickup("100bpm");
  pickups[1] = new Pickup("120bpm");
  pickups[2] = new Pickup("140bpm");
  pickups[3] = new Pickup("160bpm");
  pickups[4] = new Pickup("180bpm");
}

public void draw()
{
  background(0);

  millicounter += (millis() - previousmillis);

  if(millicounter >= mspt)
  {
    // a new tick happens
    currentTick++;
    if(currentTick >= bpmeasure * tpb)
    {
      currentTick = 0;
    }
    millicounter = millicounter - mspt;
    if(currentTick % tpb == 0)
    {
      beatAudio.rewind();
      beatAudio.play();
    }
  } 

  doPickups();

  radiation.render();

  doBalls();

  drawInfo();

  previousmillis = millis();
}

public void setBPM(int bpm)
{
  // calculate integer value for mspt
  mspt = PApplet.parseInt(60000.0f / (bpm * tpb));
  // set beats per minute
  bpminute = bpm;
}

// draws the infobar with everything in it
public void drawInfo()
{
  fill(20);
  strokeWeight(4);
  stroke(255);
  rect(2, height - infoBarHeight, fpsWidth, infoBarHeight-2);
  rect(fpsWidth + 2, height - infoBarHeight, bpmWidth, infoBarHeight-2);  
  rect(bpmWidth + fpsWidth + 2, height - infoBarHeight, width - bpmWidth - fpsWidth - 4, infoBarHeight-2);
  drawTextInfo();
  drawBeatInfo();
}

// draws the text info on frame rate and beats per minute
public void drawTextInfo()
{
  fpstext = "FPS = " + PApplet.parseInt(frameRate);
  bpmtext = "BPM = " + bpminute;
  fill(255);
  text(fpstext, 20, height - 22);
  text(bpmtext, 210, height - 22);
}

// draws the measure in the right bottom and all information in it
public void drawBeatInfo()
{
  // draw the stripes indicating the ticks
  strokeWeight(4);
  int hgap = 25;
  int startingX = fpsWidth + bpmWidth + hgap;
  int totalWidth = width - fpsWidth - bpmWidth - hgap * 2;
  int step = totalWidth / (bpmeasure * tpb);
  float infobarFraction = 0.53f;
  float vgap = infoBarHeight * ((1-infobarFraction)/2);

  for(int i = 0; i <= (bpmeasure * tpb); i++)
  {
    if(i % tpb == 0)
    {
      stroke(255);
    }
    else
    {
      stroke(120);
    }
    line(startingX + i * step, height - infoBarHeight + vgap, startingX + i * step, height - vgap);
  }

  // draw the rectangle indicating the current tick
  strokeWeight(1);
  float rectLeft = startingX + currentTick * step + step/6;
  float rectTop = height - infoBarHeight + vgap + 4;
  rect(rectLeft, rectTop, step * 2/3, infoBarHeight - 2 * vgap - 8);

  // draw the ball coloured stripes indicating all activated balls
  strokeWeight(3);
  float stripeFactor = ((infoBarHeight - 8) - 2 * vgap) / balls.length;
  for(int t = 0; t < (bpmeasure*tpb); t++)
  {
    for(int c = 0; c < balls.length; c++)
    {
      if(balls[c].on[t])
      {
        stroke(getColour(c+1));
        rectLeft = startingX + t * step + step/4;
        int h = PApplet.parseInt(rectTop + 2.0f + c * stripeFactor);
        line(rectLeft, h, rectLeft + step / 2, h);
      }
    }
  }
  strokeWeight(1);

}

// performs all actions that have to do with the balls
public void doBalls()
{
  for(int c = 0; c < balls.length; c++)
  {
    balls[c].move();
    balls[c].checkForSound();    
    balls[c].checkForPlayerHit(radiation.getOuterRadius() - 20);
    for(int i = 0; i < balls.length; i++)
    {
      if(i != c)
      {
        balls[c].checkForBallCollision(balls[i]);
      }
    }
  }
}

// performs all actions that have to do with the pickups
public void doPickups()
{
  for(int p = 0; p < pickups.length; p++)
  {
    if(pickups[p].active)
    {
      pickups[p].render();
      pickups[p].checkForPlayerHit(radiation.getOuterRadius() - 50);
    }
    else
    {
      if(pickups[p].getBPMValue() != bpminute && random(1.0f) < 0.001f)
      {
        pickups[p].activate();
      }
    }
  }
}

// returns the colour belonging to the player(number)
public int getColour(int playerNumber)
{
  switch(playerNumber)
  {
  case 1:
    return color(160, 0, 0); // red
  case 2:
    return color(60, 70, 180); // blue
  case 3:
    return color(30, 120, 15); // dark green
  case 4:
    return color(255, 120, 0); // orange
  case 5:
    return color(255, 255, 50); // yellow
  case 6:
    return color(40, 230, 230); // cyan
  case 7:
    return color(100, 0, 100); // purple
  case 8:
    return color(255, 60, 255); // pink
  case 9:
    return color(100, 255, 0); // light green
  case 10:
    return color(120, 80, 30); // brown
  default:
    return color(0, 0, 0);
  }
}

// close off everything
public void stop()
{
  // always close Minim audio classes
  for(int c = 0; c < balls.length; c++)
  {
    balls[c].audio.close();
  }
  beatAudio.close();
  super.stop();
}

class Ball
{
  float xpos;
  float ypos;
  float speed;        // in pixels per frame
  float direction;    // in radians
  float diameter;
  int colour;

  int invincCounter;                                      // time a ball will turn invincible when hit; in ms

  int[] soundCounters = new int[bpmeasure * tpb];         // how many times sound has sounded while on
  boolean[] soundSounded = new boolean[bpmeasure * tpb];  // if sound already played for current tick
  boolean[] on = new boolean[bpmeasure * tpb];            // if activated for current tick
  AudioSnippet audio;                                     //  to play the ball's sound

  Ball(String af, int c)
  {
    speed = random(4.0f) + 2.0f;
    direction = random(TWO_PI);
    diameter = random(12.0f) + 12.0f;
    reposition();
    colour = c;
    audio = Minim.loadSnippet(af);
  }

  public void reposition()
  {
    xpos = random(width - 2 * diameter) + diameter;
    ypos = random(height - 2 * diameter - infoBarHeight) + diameter;
  }
  
  public void move()
  {
    xpos = xpos + cos(direction) * speed;
    ypos = ypos + sin(direction) * speed;
    float r = 0;//diameter/2;
    if(xpos < r)
    {
      xpos = width - r;
    }
    else if(xpos > width - r)
    {
      xpos = r;
    }

    if(ypos < r)
    {
      ypos = height - infoBarHeight - r;
    }
    else if(ypos > height - infoBarHeight - r)
    {
      ypos = r;
    }
    render();
  }

  public void render()
  {
    fill(colour);
    stroke(colour);
    ellipse(xpos, ypos, diameter, diameter);
  }

  // the ball checks if it is hit by the player, and if true,
  // changes direction slightly, becomes invincible temporarily, and changes its sound pattern
  // currentRadius is the radius within which a hit will occur  
  public void checkForPlayerHit(float currentRadius)
  {
    // see if the ball is currently invincible to the player
    boolean invinc;
    if(invincCounter > 0)
    {
      invincCounter -= millis() - previousmillis;
      invinc = true;
    }
    else
    {
      invinc = false;
    }
    // determine the direct distance to player
    float distance = dist(mouseX, mouseY, xpos, ypos) - diameter/2.0f;

    if(!invinc && distance < currentRadius) 
    {
      // hit for the first time by player
      // switch sound on (or 'reset')
      on[currentTick] = true;
      soundCounters[currentTick] = 0;
      soundSounded[currentTick] = false;

      // turn invincible for one OR two OR three ticks (randomly)
      float r = random(1.0f);
      if(r < 0.333f)
      {
        invincCounter = mspt;
      }
      else if(r > 0.333f && r < 0.667f)
      {
        invincCounter = 2 * mspt;
      }
      else
      {
        invincCounter = 3 * mspt;
      }      
    }
    else if(invinc)
    {
      // ball is invincible so was already hit for the first time;
      // change direction randomly every frame
      float directionChange = random(0.6f) - 0.3f;
      direction += directionChange;
    }  
  }

  // the ball checks if it is hit by another ball, and if true,
  // exchanges direction and speed with the other one
  public boolean checkForBallCollision(Ball other)
  {
    float distance = dist(this.xpos, this.ypos, other.xpos, other.ypos);
    distance -= diameter / 2.0f;
    distance -= other.diameter / 2.0f;
    if(distance <= 0.0f) 
    {
      //exchange speed values
      float tempspeed = other.speed;
      other.speed = this.speed;
      this.speed = tempspeed;

      // exchange direction values          
      float tempdirection = other.direction;
      other.direction = this.direction;
      this.direction = tempdirection;
      
      return true;
    }
    return false;
  }

  // the ball checks if it is time to play its sound
  // also, resets values for soundSounded
  public void checkForSound()
  {
    if(on[currentTick])
    {
      // blink, also when sound sounded already in previous frame 
      // within this tick (to make it more visible)
      blink();
      // play its sound only when sound hasn't sounded already for this tick
      if(!soundSounded[currentTick])
      {
        audio.rewind();
        audio.play();
        soundSounded[currentTick] = true;
        soundCounters[currentTick]++;
        // stop making the sound if the limit is reached
        if(soundCounters[currentTick] > soundLimit)
        {
          on[currentTick] = false;
          soundCounters[currentTick] = 0;
        }
      }
    }

    // always set the soundSounded of former tick back to false
    if(currentTick == 0)
    {
      soundSounded[soundSounded.length - 1] = false;
    }
    else
    {
      soundSounded[currentTick - 1] = false;
    }
  }

  public void blink()
  {
    // show up white for one frame
    stroke(100);
    fill(0);
    ellipse(xpos, ypos, 1.6f*diameter, 1.6f*diameter);
    stroke(255);
    ellipse(xpos, ypos, 1.3f*diameter, 1.3f*diameter);
    fill(255);
    ellipse(xpos, ypos, diameter, diameter);
    redraw();
  }    
}

class Pickup
{
  PImage img;
  String imgstring;
  boolean active;
  float xpos, ypos;
  int endtime;
  int lifetime = 5000;
  
  Pickup(String imagestring)
  {
    imgstring = imagestring;
    img = loadImage(imgstring + ".png");
    active = false;
  }
  
  public void render()
  {
      if(millis() >= endtime)
      {
        deactivate();
      }
      else
      {
        image(img, xpos, ypos);
      }        
  }
  
  public void activate()
  {
    active = true;
    xpos = random(width - img.width);
    ypos = random(height - 60 - img.height);
    endtime = millis() + lifetime;
  }
  
  public void deactivate()
  {
    active = false;
  }
  
  public void checkForPlayerHit(float currentRadius)
  {
    float hw = img.width/2.0f;
    float distance = dist(mouseX, mouseY, xpos + hw, ypos + hw);
    if(distance < currentRadius) 
    {
      // player picks up the pickup; what happens depends on function
      setBPM(getBPMValue());
      deactivate();      
    }   
  }

  public int getBPMValue()
  {
    return parseInt(imgstring.substring(0, 3));
  }
}

class Radiation
{
  int nrCircles = 3;
  int[] diameters = new int[nrCircles];
  int[] graytones = new int[nrCircles];
  int dia_incr = 1;
  int gray_decr = 3;

  Radiation()
  {
    int gstep = 255/nrCircles;
    int dstep = getMaxRadius()/nrCircles;
    for(int c = 0; c < nrCircles; c++)
    {
      graytones[c] = 255 - c * gstep;
      diameters[c] = c * dstep;
    }
  }

  public void render()
  {
    for(int e = 0; e < nrCircles; e++)
    {
      diameters[e] += dia_incr;
      graytones[e] -= gray_decr;
      if(graytones[e] <= 0)
      {
        diameters[e] = 0;
        graytones[e] = 255;
      }
      stroke(graytones[e]);
      noFill();
      ellipse(mouseX, mouseY, diameters[e], diameters[e]);
    }
  }

  public int getMaxRadius()
  {
    return dia_incr * (255/gray_decr);
  }

  public int getOuterRadius()
  {
    // look for the radius belonging to the smallest circle
    // where it's gray value is above 30
    int smallestgray = 255;
    int index = -1;
    for(int i = 0; i < nrCircles; i++)
    {
      if(graytones[i] < smallestgray && graytones[i] > 30)
      {
        smallestgray = graytones[i];
        index = i;
      }
    }
    return diameters[index];
  }
}   

  static public void main(String args[]) {     PApplet.main(new String[] { "soundBalls3" });  }}