import processing.opengl.*;
import ddf.minim.*;

// fps = frames per second
// bpm = beats per minute
// bps = beats per second
// bpmeasure = beats per measure
// tps = ticks per second
// fpt = frames per tick
// tpb = ticks per beat
int tpb = 3;
int bpmeasure = 4;
int mspt;                  // how many ms every tick takes
int bpminute;              // current beats per minute (tempo)
int currentTick = 0;       // which tick in the measure we are now
int millicounter = 0;      // how many ms passed since previous tick            
int previousmillis = 0;    // how many ms passed since previous frame            
int soundLimit = 11;       // number of measures balls play their sound once triggered

String fpstext = "INIT";
String bpmtext = "INIT";
int infoBarHeight = 70;    // height of bottom info bar
int fpsWidth = 190;        // weight of part of info bar showing FPS
int bpmWidth = 223;        // weight of part of info bar showing BPM

// the radiation radiated by the player
Radiation radiation = new Radiation();
// array of balls in the game (maximum of 10)
Ball[] balls = new Ball[10];
// array of pickups in the game
Pickup[] pickups = new Pickup[5];

AudioSnippet beatAudio;

void setup()
{
  size(900, 650, OPENGL); // 900, 650
  noFill();
  smooth();
  noCursor();  

  // try to achieve 60 frames per second and start with 100 bpm
  frameRate(60);
  setBPM(100);

  textFont(loadFont("Pinball-Data-48.vlw"));
  textAlign(LEFT);
  rectMode(CORNER);
  strokeWeight(1);

  // initialize sound
  Minim.start(this);
  beatAudio = Minim.loadSnippet("button2.wav");
  // sound indicating beat set off by setting it to -80 dB
  beatAudio.setGain(-80);    //-6

  // spawn all Balls
  String audioFile;
  for(int c = 0; c < balls.length; c++)
  {
    audioFile = "c" + (c + 1) + ".wav";
    balls[c] = new Ball(audioFile, getColour(c + 1));
    boolean collision = true;
    while(collision)
    {
      for(int i = 0; i < c; i++)
      {
        if(i != c)
        {
          if(balls[c].checkForBallCollision(balls[i]))
          {
            collision = true;
            balls[c].reposition();
            break;
          }
        }
      }
      collision = false;
    }
  }

  // prepare all Pickups
  pickups[0] = new Pickup("100bpm");
  pickups[1] = new Pickup("120bpm");
  pickups[2] = new Pickup("140bpm");
  pickups[3] = new Pickup("160bpm");
  pickups[4] = new Pickup("180bpm");
}

void draw()
{
  background(0);

  millicounter += (millis() - previousmillis);

  if(millicounter >= mspt)
  {
    // a new tick happens
    currentTick++;
    if(currentTick >= bpmeasure * tpb)
    {
      currentTick = 0;
    }
    millicounter = millicounter - mspt;
    if(currentTick % tpb == 0)
    {
      beatAudio.rewind();
      beatAudio.play();
    }
  } 

  doPickups();

  radiation.render();

  doBalls();

  drawInfo();

  previousmillis = millis();
}

void setBPM(int bpm)
{
  // calculate integer value for mspt
  mspt = int(60000.0 / (bpm * tpb));
  // set beats per minute
  bpminute = bpm;
}

// draws the infobar with everything in it
void drawInfo()
{
  fill(20);
  strokeWeight(4);
  stroke(255);
  rect(2, height - infoBarHeight, fpsWidth, infoBarHeight-2);
  rect(fpsWidth + 2, height - infoBarHeight, bpmWidth, infoBarHeight-2);  
  rect(bpmWidth + fpsWidth + 2, height - infoBarHeight, width - bpmWidth - fpsWidth - 4, infoBarHeight-2);
  drawTextInfo();
  drawBeatInfo();
}

// draws the text info on frame rate and beats per minute
void drawTextInfo()
{
  fpstext = "FPS = " + int(frameRate);
  bpmtext = "BPM = " + bpminute;
  fill(255);
  text(fpstext, 20, height - 22);
  text(bpmtext, 210, height - 22);
}

// draws the measure in the right bottom and all information in it
void drawBeatInfo()
{
  // draw the stripes indicating the ticks
  strokeWeight(4);
  int hgap = 25;
  int startingX = fpsWidth + bpmWidth + hgap;
  int totalWidth = width - fpsWidth - bpmWidth - hgap * 2;
  int step = totalWidth / (bpmeasure * tpb);
  float infobarFraction = 0.53;
  float vgap = infoBarHeight * ((1-infobarFraction)/2);

  for(int i = 0; i <= (bpmeasure * tpb); i++)
  {
    if(i % tpb == 0)
    {
      stroke(255);
    }
    else
    {
      stroke(120);
    }
    line(startingX + i * step, height - infoBarHeight + vgap, startingX + i * step, height - vgap);
  }

  // draw the rectangle indicating the current tick
  strokeWeight(1);
  float rectLeft = startingX + currentTick * step + step/6;
  float rectTop = height - infoBarHeight + vgap + 4;
  rect(rectLeft, rectTop, step * 2/3, infoBarHeight - 2 * vgap - 8);

  // draw the ball coloured stripes indicating all activated balls
  strokeWeight(3);
  float stripeFactor = ((infoBarHeight - 8) - 2 * vgap) / balls.length;
  for(int t = 0; t < (bpmeasure*tpb); t++)
  {
    for(int c = 0; c < balls.length; c++)
    {
      if(balls[c].on[t])
      {
        stroke(getColour(c+1));
        rectLeft = startingX + t * step + step/4;
        int h = int(rectTop + 2.0 + c * stripeFactor);
        line(rectLeft, h, rectLeft + step / 2, h);
      }
    }
  }
  strokeWeight(1);

}

// performs all actions that have to do with the balls
void doBalls()
{
  for(int c = 0; c < balls.length; c++)
  {
    balls[c].move();
    balls[c].checkForSound();    
    balls[c].checkForPlayerHit(radiation.getOuterRadius() - 20);
    for(int i = 0; i < balls.length; i++)
    {
      if(i != c)
      {
        balls[c].checkForBallCollision(balls[i]);
      }
    }
  }
}

// performs all actions that have to do with the pickups
void doPickups()
{
  for(int p = 0; p < pickups.length; p++)
  {
    if(pickups[p].active)
    {
      pickups[p].render();
      pickups[p].checkForPlayerHit(radiation.getOuterRadius() - 50);
    }
    else
    {
      if(pickups[p].getBPMValue() != bpminute && random(1.0) < 0.001)
      {
        pickups[p].activate();
      }
    }
  }
}

// returns the colour belonging to the player(number)
color getColour(int playerNumber)
{
  switch(playerNumber)
  {
  case 1:
    return color(160, 0, 0); // red
  case 2:
    return color(60, 70, 180); // blue
  case 3:
    return color(30, 120, 15); // dark green
  case 4:
    return color(255, 120, 0); // orange
  case 5:
    return color(255, 255, 50); // yellow
  case 6:
    return color(40, 230, 230); // cyan
  case 7:
    return color(100, 0, 100); // purple
  case 8:
    return color(255, 60, 255); // pink
  case 9:
    return color(100, 255, 0); // light green
  case 10:
    return color(120, 80, 30); // brown
  default:
    return color(0, 0, 0);
  }
}

// close off everything
void stop()
{
  // always close Minim audio classes
  for(int c = 0; c < balls.length; c++)
  {
    balls[c].audio.close();
  }
  beatAudio.close();
  super.stop();
}
